package be.iminds.ilabt.jfed.experimenter_gui.slice.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Proxy;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.ProxyServiceUtil;
import javafx.util.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 17:58
 */
public class TestLinksJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(TestLinksJob.class);

    private final GeniUserProvider geniUserProvider;
    private final JFedPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final ProxyServiceUtil proxyServiceUtil;

    private final List<FXRspecNode> failedNodes = Collections.synchronizedList(new ArrayList<>());

    TestLinksJob(Experiment experiment, HighLevelTaskFactory hltf, TaskThread tt,
                 GeniUserProvider geniUserProvider, JFedPreferences jFedPreferences,
                 ProxyPreferencesManager proxyPreferencesManager, ProxyServiceUtil proxyServiceUtil) {
        super("Test connectivity to resources", experiment, hltf, tt);
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.proxyServiceUtil = proxyServiceUtil;
    }


    @Override
    public Boolean execute() throws Exception {
        LOG.trace("Starting connectivity tests to existing resources in experiment {}", experiment.getName());

        final TestLinksExperimentPartsState tcs = new TestLinksExperimentPartsState();
        setAndRunState(tcs);

        return tcs.getStatus() == ExperimentTaskStatus.SUCCESS;
    }

    public class TestLinksExperimentPartsState extends SlicedState<TestLinksExperimentPartStateSlice> {

        private final List<TestLinksExperimentPartStateSlice> slices;

        protected TestLinksExperimentPartsState() {
            super("Test connectivity to experiment parts");

            slices = experiment.getPartsStream()
                    .filter(part -> part instanceof SfaExperimentPart) //only SFA parts are accesible for the moment
                    .map(TestLinksExperimentPartStateSlice::new)
                    .collect(Collectors.toList());
        }

        @Override
        public Collection<TestLinksExperimentPartStateSlice> getSlices() {
            return slices;
        }

        public List<FXRspecNode> getFailedNodes() {
            return failedNodes;
        }
    }


    public class TestLinksExperimentPartStateSlice extends ExperimentPartStateSlice {
        private TestLinksExperimentPartStateSlice(ExperimentPart experimentPart) {
            super(TestLinksJob.this, experimentPart);
        }

        @Override
        public String getName() {
            return "Testing connectivity to resources in " + experimentPart.getName();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            FindTestableResourcesState findTestableResourcesState = new FindTestableResourcesState(experimentPart);

            setAndRunState(findTestableResourcesState);

            if (findTestableResourcesState.getStatus() == ExperimentTaskStatus.SUCCESS &&
                    !findTestableResourcesState.getNodesToTest().isEmpty()) {
                final TestLinksOfExperimentPartState tcs =
                        new TestLinksOfExperimentPartState(experimentPart,
                                findTestableResourcesState.getNodesToTest());
                setAndRunState(tcs);

                Map<String, FXRspecNode> failedNodesByClientId =
                        tcs.linkReports.stream()
                                .filter(c -> c.getSummary() != LinkTestResult.SUCCESS)
                                .map(LinkTestReport::getFrom)
                                .collect(Collectors.toMap(FXRspecNode::getClientId, Function.identity(), (a, b)->a));
                failedNodes.addAll(failedNodesByClientId.values());

                return tcs.getStatus();
            } else {
                LOG.debug("Skipping: no testable resources found");
                return ExperimentTaskStatus.SUCCESS;
            }
        }
    }

    /**
     * The connectivity-test can begin if, for each sliver of this experimentPart if:
     * - the status of the sliver is READY
     * - the manifest of the sliver is available
     * <p/>
     * If the authority has one or more nodes, it must have at least one loginService available
     * (we assume that if one loginService is available, it will be available for all loginServices in that sliver)
     */
    public class FindTestableResourcesState extends State {

        private final ExperimentPart experimentPart;

        private Collection<FXRspecNode> nodesToTest = null;

        private FindTestableResourcesState(ExperimentPart experimentPart) {
            super(String.format("Determining testable resources on %s", experimentPart.getName()));
            this.experimentPart = experimentPart;
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            //use a map to make sure that only one rspecnode per id is collected
            Map<String, FXRspecNode> nodesOfAuthorityWithLoginService = new HashMap<>();
            Set<String> nodeIdsWithoutLoginService = new HashSet<>();

            if (!(experimentPart instanceof SfaExperimentPart)) {
                LOG.debug("Currently no support for connectivity testing of {}", experimentPart.getClass().getName());
                nodesToTest = Collections.emptyList();

                return ExperimentTaskStatus.WARNING;
            }

            SfaExperimentPart sfaExperimentPart = (SfaExperimentPart) experimentPart;

            assert sfaExperimentPart.getSlivers() != null;
            for (Sliver sliver : sfaExperimentPart.getSlivers()) {
                //skip fake slivers
                if(sliver.isFake())
                    continue;

                //skip virtual slivers (ex. vlan-slivers on ExoGENI)
                if(sliver.isVirtual())
                    continue;

                //check if sliver is ready
                if (sliver.getStatus().getGlobalStatus() != StatusDetails.SliverStatus.READY) {
                    if (LOG.isTraceEnabled())
                        LOG.trace("Cannot start connectivity test to {} as sliver {} is not READY", experimentPart.getName(),
                                sliver.getUrn());

                    return ExperimentTaskStatus.WARNING;
                }

                //check if manifest is available
                if (sliver.getManifestRspec() == null) {
                    if (LOG.isTraceEnabled())
                        LOG.trace("Cannot start connectivity test to {} as manifest is not available for sliver {}",
                                experimentPart.getName(), sliver.getUrn());

                    return ExperimentTaskStatus.WARNING;
                } else {
                    //only add node if it has a loginService, and hasn't been added yet to the map
                    FXModelRspec modelRspec = sliver.getManifestRspec().getModelRspec(ModelRspecType.FX);
                    assert modelRspec != null;
                    //noinspection ConstantConditions
                    Collection<FXRspecNode> nodes = modelRspec.getNodesByAuthority(sfaExperimentPart.getConnectSfaAuthority());
                    assert nodes != null;
                    for (FXRspecNode node : nodes) {
                        if (!node.getLoginServices().isEmpty()) {
                            if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId())) {
                                nodesOfAuthorityWithLoginService.put(node.getUniqueId(), node);
                                nodeIdsWithoutLoginService.remove(node.getUniqueId()); //remove it from the nodeIdsWithoutLoginService
                            }
                        } else {
                            //check if we haven't found a loginService for this node before, and add it to nodeIdsWithoutLoginService otherwise
                            if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId()))
                                nodeIdsWithoutLoginService.add(node.getUniqueId());
                        }
                    }
                }
            }

            if (!nodesOfAuthorityWithLoginService.isEmpty()) {
                nodesToTest = nodesOfAuthorityWithLoginService.values();
            } else if (!nodeIdsWithoutLoginService.isEmpty()) {
                if (LOG.isTraceEnabled())
                    LOG.trace("Cannot start connectivity test for authority {} as no nodes with loginServices are available",
                            experimentPart.getName());
                nodesToTest = Collections.emptyList(); //we didn't find a loginService
            } else {
                //start the connectivity test
                if (LOG.isTraceEnabled())
                    LOG.trace("No connectivity test will be performed for authority {}, as it isn't responsible for any nodes",
                            experimentPart.getName());
                nodesToTest = Collections.emptyList();   //the authority isn't responsible for a node
            }

            return ExperimentTaskStatus.SUCCESS;
        }

        public Collection<FXRspecNode> getNodesToTest() {
            return nodesToTest;
        }
    }

    public enum LinkTestResult {
        SUCCESS, PARTIAL_FAILURE, FAILURE;
        boolean isWorseThan(@Nonnull LinkTestResult other) {
            switch (this) {
                case SUCCESS: return false;
                case PARTIAL_FAILURE: return other == SUCCESS;
                case FAILURE: return other != FAILURE;
                default: throw new RuntimeException("Unhandled LinkTestResult"+other);
            }
        }
    }
    public static class LinkTestReport {
        @Nonnull private final FXRspecNode from;
        @Nonnull private final FXRspecNode to;

        private final LinkTestResult summary;

        private final int expectedLinkSpeed_mbps;

        private final boolean connectivityOk;
        private final boolean linkSpeedSettingOk;
        private final boolean throughputOk;
        private final boolean lossOk;

        private final int actualLinkSpeedSetting_mbps;
        private final double actualThroughput_mbps;
        private final double actualLoss_percent;

        public LinkTestReport(@Nonnull FXRspecNode from, @Nonnull FXRspecNode to,
                              int expectedLinkSpeed_mbps, boolean connectivityOk,
                              double actualThroughput_mbps, int actualLinkSpeedSetting_mbps,
                              double actualLoss_percent) {
            this.from = from;
            this.to = to;
            this.expectedLinkSpeed_mbps = expectedLinkSpeed_mbps;
            this.connectivityOk = connectivityOk;
            this.actualThroughput_mbps = actualThroughput_mbps;
            this.actualLinkSpeedSetting_mbps = actualLinkSpeedSetting_mbps;
            this.actualLoss_percent = actualLoss_percent;

            this.linkSpeedSettingOk = actualLinkSpeedSetting_mbps == expectedLinkSpeed_mbps;
            this.lossOk = actualLoss_percent <= 0.001;
            //low expectations about actual throughput:
            if (expectedLinkSpeed_mbps == 10)
                this.throughputOk = actualThroughput_mbps > 2.0;
            else if (expectedLinkSpeed_mbps == 100)
                this.throughputOk = actualThroughput_mbps > 20.0;
            else if (expectedLinkSpeed_mbps == 1000)
                this.throughputOk = actualThroughput_mbps > 200.0;
            else
                this.throughputOk = actualThroughput_mbps > (expectedLinkSpeed_mbps / 10.0);

            LinkTestResult summary = LinkTestResult.FAILURE;
            if (connectivityOk && linkSpeedSettingOk && this.throughputOk && lossOk)
                summary = LinkTestResult.SUCCESS;
            if (connectivityOk || this.throughputOk || lossOk)
                summary = LinkTestResult.PARTIAL_FAILURE;
            this.summary = summary;
        }

        @Nonnull
        public FXRspecNode getFrom() {
            return from;
        }

        @Nonnull
        public FXRspecNode getTo() {
            return to;
        }

        public LinkTestResult getSummary() {
            return summary;
        }

        public int getExpectedLinkSpeed_mbps() {
            return expectedLinkSpeed_mbps;
        }

        public boolean isConnectivityOk() {
            return connectivityOk;
        }

        public boolean isLinkSpeedSettingOk() {
            return linkSpeedSettingOk;
        }

        public boolean isThroughputOk() {
            return throughputOk;
        }

        public boolean isLossOk() {
            return lossOk;
        }

        public int getActualLinkSpeedSetting_mbps() {
            return actualLinkSpeedSetting_mbps;
        }

        public double getActualThroughput_mbps() {
            return actualThroughput_mbps;
        }

        public double getActualLoss_percent() {
            return actualLoss_percent;
        }

        @Nullable
        public static LinkTestReport worst(@Nonnull Collection<LinkTestReport> reports) {
            if (reports.isEmpty()) return null;
            LinkTestReport res = reports.iterator().next();
            for (LinkTestReport cur: reports) {
                if (cur.summary.isWorseThan(res.summary))
                    res = cur;
            }
            return res;
        }
    }

    public class TestLinksOfExperimentPartState extends State {

        private final ExperimentPart experimentPart;
        private final Collection<FXRspecNode> nodesToTest;

        private final List<LinkTestReport> linkReports =
                new ArrayList<>();

        private TestLinksOfExperimentPartState(ExperimentPart experimentPart,
                                               Collection<FXRspecNode> nodesToTest) {
            super(String.format("Testing links of %s", experimentPart.getName()));
            this.experimentPart = experimentPart;
            this.nodesToTest = nodesToTest;
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            assert (!nodesToTest.isEmpty());
            assert experimentPart instanceof SfaExperimentPart;
            assert ((SfaExperimentPart) experimentPart).getSlivers() != null;

            for (Sliver sliver : ((SfaExperimentPart) experimentPart).getSlivers()) {
                assert (sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.READY);
                assert (sliver.isFake() || sliver.getManifestRspec() != null);
            }


            JFedConnection.ProxyInfo proxyInfo = null;
            if (Objects.equals(jFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH), ProxyPreferencesManager.ALWAYS_USE_PROXY)) {
                GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                assert geniUser.getUserAuthorityServer() != null;
                //noinspection ConstantConditions
                List<Proxy> userAuthProxyInfos = geniUser.getUserAuthorityServer().getTestbed().getProxies();
                proxyInfo =
                        proxyPreferencesManager.getSshProxySettings(userAuthProxyInfos, geniUser);

                if (proxyInfo == null) {
                    LOG.warn("Connectivity tests to nodes on authority {} are skipped as SSH-proxy is active but proxy info was not found." +
                            " (this is a jFed bug)", experimentPart.getName());


                    updateMessage("Skipped because proxy for SSH-connections is active but proxy info was not found. (this is a jFed bug)");
                    return ExperimentTaskStatus.FAILED;
                }
            }


            LOG.debug("Starting nodes connectivity test for {} nodes on {}", nodesToTest.size(), experimentPart.getName());

            Iterator<FXRspecNode> it = nodesToTest.iterator();
            while (it.hasNext()) {
                FXRspecNode node = it.next();

                if (node.getLoginServices().isEmpty()) {
                    LOG.debug("Removing node {} from test because it doesn't have any loginServices available", node.getUniqueId());
                    it.remove();
                }
            }

            LinkTestReport allTimeWorst = null;
            for (FXRspecNode node : nodesToTest) {
                List<LinkTestReport> result = testLinksOfNode(node, proxyInfo);

                LinkTestReport curWorst = LinkTestReport.worst(result);
                if (curWorst != null) {
                    if (allTimeWorst == null) {
                        allTimeWorst = curWorst;
                    } else {
                        allTimeWorst = LinkTestReport.worst(Arrays.asList(curWorst, allTimeWorst));
                    }
                }

                linkReports.addAll(result);
            }

            if (allTimeWorst == null || allTimeWorst.getSummary() == LinkTestResult.SUCCESS) {
                updateMessage("");
                return ExperimentTaskStatus.SUCCESS;
            } else {
                LOG.trace("Link test failed.");
                updateMessage("Link test failed.");
                return ExperimentTaskStatus.FAILED;
            }
        }

        private List<LinkTestReport> testLinksOfNode(FXRspecNode node, JFedConnection.ProxyInfo proxyInfo) {
            LoginService loginService = node.getLoginServices().get(0);

            LOG.trace("Going to try to connect to {} by using loginService {}:{}",
                    node.getUniqueId(), loginService.getHostname(), loginService.getPort());


            JFedConnection.ProxyInfo testbedProxyInfo = proxyServiceUtil.findTestbedProxy(node, loginService);
            if (testbedProxyInfo != null)
                LOG.trace("Trying to make connection to {} with testbed proxy {}", node.getUniqueId(), testbedProxyInfo);

            if (proxyInfo != null && testbedProxyInfo != null) {
                LOG.warn("jFed has proxy currently enabled. However, connectivity test will only use testbed proxy {} while trying to connect to {}!", testbedProxyInfo, node.getUniqueId());
            }

            String nodeClientId = node.getClientId();
            List<LinkTestReport> res = new ArrayList<>();

            //TODO: start thoughput_server_tcp.py on all other nodes

            List<Pair<FXRspecNode, String>> targetIPsByNode = new ArrayList<>();
            for (FXRspecInterface iface: node.getInterfaces()) {
                String ifaceClientId = iface.getClientId();
                String command = String.format("/tmp/link_tester.py %s %s", nodeClientId, ifaceClientId);
                for (FXRspecInterface otherIface: iface.getLink().getInterfaces()) {
                    //noinspection ObjectEquality
                    if (otherIface == iface)
                        continue;
            //TODO: start
                    FXRspecNode targetNode = otherIface.getNode();
                    if (otherIface.getIpAddresses() != null && !otherIface.getIpAddresses().isEmpty()) {
                        String ip = otherIface.getIpAddresses().get(0).getAddress();
                        targetIPsByNode.add(new Pair<>(targetNode, ip));
                    }

                    String targetNodeId = targetNode.getClientId();

                    command += " " + targetNodeId;


                    //TODO Use real test data instead of these dummy values
                    LinkTestReport rep;
                    if (Math.random() < 0.5)
                        rep = new LinkTestReport(node, targetNode,
                                1000, Math.random() < 0.5,
                                Math.random() * 99.2, 100,
                                Math.random() * 5.0);
                    else
                        rep = new LinkTestReport(node, targetNode,
                                1000, true,
                                500.0 + (Math.random()*450.0), 1000,
                                0);

                    res.add(rep);
                }


                //TODO: execute command on node, and process result to make LinkTestReports
                //command

                /*
                   ~$ python link_tester.py node0 node0:if0 node1 node2
                    [{
                        "target": "node1",
                        "ping": true,
                        "measured_bw": 932.424213606,
                        "node_client_id": "node0",
                        "iface_dev": "eth3",
                        "iface_speed": "1000",
                        "iface_client_id": "node0:if0"
                    }, {
                        "target": "node2",
                        "ping": true,
                        "measured_bw": 929.627972989,
                        "node_client_id": "node0",
                        "iface_dev": "eth3",
                        "iface_speed": "1000",
                        "iface_client_id": "node0:if0"
                    }]
                */
            }

            //TODO: stop thoughput_server_tcp.py on all other nodes

            return res;
        }

        public Collection<FXRspecNode> getNodesToTest() {
            return nodesToTest;
        }

        public List<LinkTestReport> getLinkReports() {
            return linkReports;
        }
    }
}
